/*
 * (C) Copyright IBM Corp. 1998  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.parser;

import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;

import org.xml.sax.EntityHandler;
import org.xml.sax.DocumentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.HandlerBase;
import org.w3c.dom.Notation;

/**
 * <p>This is a <a href="http://www.microstar.com/XML/SAX/">SAX (Simple API for XML)</a>
 * driver for IBM XML for Java.</p>
 *
 * <ul>
 *   <li><code>EntityHandler#changeEntity()</code> has never called.
 * </ul>
 *
 * @version 19980206
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 */
public class SAXDriver extends DefaultElementFactory implements org.xml.sax.Parser, org.xml.sax.AttributeMap, StartTagHandler, ErrorListener, StreamProducer {

    static HandlerBase s_defaulthandler = new HandlerBase();

    EntityHandler m_entityhandler = s_defaulthandler;
                                                // sax.Parser
    public void setEntityHandler(EntityHandler handler) {
        m_entityhandler = handler;
    }

    DocumentHandler m_documenthandler = s_defaulthandler;
                                                // sax.Parser
    public void setDocumentHandler(DocumentHandler handler) {
        m_documenthandler = handler;
    }

    ErrorHandler m_errorhandler = s_defaulthandler;
                                                // sax.Parser
    public void setErrorHandler(ErrorHandler handler) {
        m_errorhandler = handler;
    }

    String m_currentsysid;
                                                // sax.Parser
    public void parse(String publicid, String systemid) throws java.lang.Exception {
        m_currentsysid = systemid = m_entityhandler.resolveEntity("[document]", publicid, systemid);
        m_documenthandler.startDocument();
        m_entityhandler.changeEntity(systemid);

        try {
            URL u = new URL(systemid);
            Parser p = new Parser(systemid, this, this);
            p.setElementFactory(this);
            p.setStartTagHandler(this);
            p.readStream(u.openStream());
        } catch (Exception e) {
            m_errorhandler.fatal(e.getMessage(), "[unknown]", -1, -1);
        }
        m_documenthandler.endDocument();
    }
                                                // parser.StreamProducer
    public java.io.InputStream getInputStream(String name, String pubid, String sysid) throws java.io.IOException {
        if (null == sysid)  return null;
        java.io.InputStream is = null;
        try {
            URL u = new URL(m_entityhandler.resolveEntity(name, pubid, sysid));
            is = u.openStream();
        } catch (java.io.IOException e0) {
            is = null;
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
        return is;
    }
                                                // parser.ErrorListener
    public void error(String fname, int lineno, int charoff, Object key, String mes) {
        try {
            if (key instanceof String && ((String)key).startsWith("W_")) {
                m_errorhandler.warning(mes, m_currentsysid, lineno, charoff);
            } else
                m_errorhandler.fatal(mes, m_currentsysid, lineno, charoff);
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
    }

                                                // parser.DefaultElementFacotry
    DTD m_dtd = null;
    public DTD createDTD(String root, ExternalID eid) {
        try {
            if (null == eid) {
                m_documenthandler.doctype(root, null, null);
            } else {
                m_documenthandler.doctype(root, eid.getPubidLiteral(), eid.getSystemLiteral());
            }
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
        return m_dtd = super.createDTD(root, eid);
    }
                                                // parser.StartTagHandler
    Hashtable m_attributes = new Hashtable();
    public void handleStartTag(TXElement el, boolean empty) {
        TXAttribute[] aa = el.getAttributeArray();
        for (int i = 0;  i < aa.length;  i ++)
            m_attributes.put(aa[i].getQName(), aa[i]);
        try {
            m_documenthandler.startElement(el.getQName(), this);
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
        m_attributes.clear();
    }
                                                // parser.DefaultElementFactory
    public void ripenElement(TXElement el) {
        try {
            m_documenthandler.endElement(el.getQName());
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
    }
                                                // parser.DefaultElementFactory
    public TXText createText(String data, boolean ignorable) {
        try {
            char[] ac = data.toCharArray();
            if (ignorable)
                m_documenthandler.ignorable(ac, 0, ac.length);
            else
                m_documenthandler.characters(ac, 0, ac.length);
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
        return super.createText(data, ignorable);
    }
                                                // parser.DefaultElementFactory
    public TXCDATASection createCDATA(String data) {
        try {
            char[] ac = data.toCharArray();
            m_documenthandler.characters(ac, 0, ac.length);
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
        return super.createCDATA(data);
    }
                                                // parser.DefaultElementFactory
    public TXPI createPI(String name, String data) {
        try {
            m_documenthandler.processingInstruction(name, data);
        } catch (Exception e) {
            e.printStackTrace();                // ???
        }
        return super.createPI(name, data);
    }

                                                // sax.AttributeMap
    public Enumeration getAttributeNames() {
        return m_attributes.keys();
    }
                                                // sax.AttributeMap
    public String getValue(String attributeName) {
        return ((TXAttribute)m_attributes.get(attributeName)).getStringValue();
    }
                                                // sax.AttributeMap
                                                // What value should it return for ENTITIES attribute?
    public boolean isEntity(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return false;
        return TXAttribute.T_ENTITY == a.getType();
    }
                                                // sax.AttributeMap
    public boolean isNotation(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return false;
        return TXAttribute.T_NOTATION == a.getType();
    }
                                                // sax.AttributeMap
    public boolean isId(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return false;
        return TXAttribute.T_ID == a.getType();
    }
                                                // sax.AttributeMap
                                                // What value should it return for IDREFS attribute?
    public boolean isIdref(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return false;
        return TXAttribute.T_IDREF == a.getType();
    }
                                                // sax.AttributeMap
    public String getEntityPublicID(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return null;
        if (TXAttribute.T_ENTITY != a.getType())  return null;

        EntityValue ev = m_dtd.m_entity.refer(a.getTypedValue()[0]);
        if (null != ev && ev.isExternal() && ev.isNDATA()) {
            return ev.getExternalID().getPubidLiteral();
        }
        return null;
    }
                                                // sax.AttributeMap
    public String getEntitySystemID(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return null;
        if (TXAttribute.T_ENTITY != a.getType())  return null;

        EntityValue ev = m_dtd.m_entity.refer(a.getTypedValue()[0]);
        if (null != ev && ev.isExternal() && ev.isNDATA()) {
            return ev.getExternalID().getSystemLiteral();
        }
        return null;
    }
                                                // sax.AttributeMap
    public String getNotationName(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return null;
        if (TXAttribute.T_NOTATION != a.getType())  return null;
        return a.getTypedValue()[0];
    }
                                                // sax.AttributeMap
    public String getNotationPublicID(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return null;
        if (TXAttribute.T_NOTATION != a.getType())  return null;
        Notation no = m_dtd.getNotation(a.getTypedValue()[0]);
        if (null == no)  return null;
        return no.getPublicIdentifier();
    }
                                                // sax.AttributeMap
    public String getNotationSystemID(String aname) {
        TXAttribute a = (TXAttribute)m_attributes.get(aname);
        if (null == a)  return null;
        if (TXAttribute.T_NOTATION != a.getType())  return null;
        Notation no = m_dtd.getNotation(a.getTypedValue()[0]);
        if (null == no)  return null;
        return no.getSystemIdentifier();
    }
}
